@@ -4,6 +4,7 @@ from urllib.parse import quote
 import networkx as nx
 from networkx.readwrite import json_graph
 from erdiagram.NodeType import NodeType
+import json
 
 class ER:
     '''
@@ -117,7 +118,7 @@ class ER:
             inverseDirection(bool): if "directed" and "inverseDirection" are both true, the arrow is drawn at the "from" end
         '''
         if self.debug: 
-            print(f">> adding edge: {fromNodeLabel} <---[{edgeLabel}]---> {toNodeLabel}")
+            print(f"   >> adding edge: {fromNodeLabel} <---[{edgeLabel}]---> {toNodeLabel}")
         if self.get_graph().has_edge(fromNodeLabel, toNodeLabel) or self.get_graph().has_edge(toNodeLabel, fromNodeLabel) :
             if self.debug:
                 print(f">> trying to add existing edge, nothing to be done.")
@@ -131,6 +132,7 @@ class ER:
         isPK=False, 
         isMultiple=False, 
         isWeak=False, 
+        isComposed=False,
         composedOf=[]):
         '''
         add attribute to graphML graph
@@ -143,17 +145,23 @@ class ER:
             composedOf(list): list of attributes this attribute is built of
         '''
 
-        attrLabel = f'{parentLabel}.{attrLabel}'
+        directLabel = f'{parentLabel}.{attrLabel}'
         if self.debug: 
             print(f">> adding attribute: {attrLabel}")
+        if not isComposed:
+            nodeType = str(NodeType.ATTRIBUTE)
+        else:
+            nodeType = str(NodeType.COMPOSED_ATTRIBUTE)
 
-        self.get_graph().add_node(attrLabel, 
-            label = attrLabel,
+        self.get_graph().add_node(directLabel, 
+            label = directLabel,
+            attrLabel = attrLabel,
+            parentLabel = parentLabel,
             isPK = isPK,
             isMultiple = isMultiple,
             isWeak = isWeak,
-            composedOf = str(composedOf),
-            nodeType= str(NodeType.ATTRIBUTE)
+            composedOf = json.dumps(composedOf),
+            nodeType = nodeType
         )
 
     def __add_graphml_relation(self, label, fromNodeLabel, toNodeLabel, fromEdgeLabel, toEdgeLabel, isWeak=False):
@@ -173,6 +181,7 @@ class ER:
         self.get_graph().add_node(relationLabel, 
             label = relationLabel,
             relation = f"{fromNodeLabel}<-[{fromEdgeLabel}]--[{toEdgeLabel}]->{toNodeLabel}",
+            relationLabel = label,
             relationFrom = fromNodeLabel,
             relationTo = toNodeLabel,
             fromEdgeLabel = fromEdgeLabel,
@@ -187,10 +196,10 @@ class ER:
     def __add_graphml_is_a(self, superClassLabel, superLabel, subLabel, isDisjunct=False, subClasses=[]):
         if isDisjunct:
             isALabel = f"{superClassLabel}.isA.{subClasses}"
-            relation = f"{superClassLabel}->[{superLabel}]->[isA]->[{subLabel}]->{subClasses}",
+            relation = f"{superClassLabel}->[{superLabel}]->[isA]->[{subLabel}]->{subClasses}"
         else:
             isALabel = f"{superClassLabel}.isA.{subClasses}"
-            relation = f"{superClassLabel}<-[{superLabel}]<-[isA]<-[{subLabel}]<-{subClasses}",
+            relation = f"{superClassLabel}<-[{superLabel}]<-[isA]<-[{subLabel}]<-{subClasses}"
         if self.debug: 
             print(f">> adding relation: {isALabel}")
         
@@ -201,7 +210,7 @@ class ER:
             superLabel = superLabel,
             subLabel = subLabel,
             isDisjunct = isDisjunct,
-            subClasses = str(subClasses),
+            subClasses = json.dumps(subClasses),
             nodeType = str(NodeType.IS_A)
         )
 
@@ -330,8 +339,11 @@ class ER:
                 print(f">> node not found, adding {nodeLabel}")
             self.add_node(nodeLabel)
 
+        if isinstance(composedOf, str):
+            composedOf = json.loads(composedOf)
+
         # add new node to graphML graph
-        self.__add_graphml_attr(nodeLabel, attrLabel, isPK, isMultiple, isWeak, composedOf)
+        self.__add_graphml_attr(nodeLabel, attrLabel, isPK, isMultiple = isMultiple, isWeak = isWeak, composedOf = composedOf, isComposed=False)
 
         # add new edge to graphML graph connecting the parent node to this new attribute
         self.__add_graphml_edge(nodeLabel, fullAttrLabel)
@@ -349,10 +361,10 @@ class ER:
             fullSubLabel = f'{nodeLabel}.{attrLabel}.{subAttrLabel}'
             
             if self.debug:
-                print(f">> >> adding sublabel {fullSubLabel} ")
+                print(f"   >> >> adding sublabel {fullSubLabel} ")
 
             # add to graphML graph
-            self.__add_graphml_attr(f'{nodeLabel}.{attrLabel}', subAttrLabel, isPK, isMultiple, isWeak)
+            self.__add_graphml_attr(f'{nodeLabel}.{attrLabel}', subAttrLabel, isPK, isMultiple, isWeak, isComposed=True)
             self.__add_graphml_edge(f'{nodeLabel}.{attrLabel}', fullSubLabel)
 
             # add to graphViz graph
@@ -395,11 +407,14 @@ class ER:
             isDisjunct(bool): Are the elements of this relation disjunct?
         '''
         # Add "X is a Y" relation - a green inverted triangle from a superclass to multiple subclasses
+        print([type(subclassParam).__name__, subclassParam])
         if not isinstance(subclassParam, list):
             subClasses = [subclassParam]
         else:
             subClasses = subclassParam
 
+        print([type(subClasses).__name__, subClasses])
+
         if not self.has_node(superClassLabel):
             self.add_node(superClassLabel)
         for i, subClassLabel in enumerate(subClasses):
@@ -433,10 +448,103 @@ class ER:
     def get_graph(self):
         return self.graph
 
+    def get_default_scores(self):
+        return {
+            'missing_node': 1,
+            'missing_property': {
+                str(NodeType.NODE): 0.5,
+                str(NodeType.ATTRIBUTE): 0.25,
+                str(NodeType.RELATION): 0.5,
+                str(NodeType.IS_A): 0.25
+            }
+        }
+        
+
+    def _add_obj_node(self, obj):
+        self.add_node(obj['label'], isMultiple = obj['isMultiple'], isWeak = obj['isWeak'])
+        if self.debug:
+            print(f"    ✓  added node {obj['label']}:")
+            print(f"       isMultiple = {obj['isMultiple']}, isWeak = {obj['isWeak']}")
+    
+    def _add_obj_attr(self, obj):
+        self.add_attribute(
+            obj['parentLabel'], obj['attrLabel'], 
+            isPK = obj['isPK'], isMultiple = obj['isMultiple'], isWeak = obj['isWeak'], 
+            composedOf = obj['composedOf']
+        )
+        if self.debug:
+            print(f"    ✓  added attribute {obj['attrLabel']} to node {obj['parentLabel']}:")
+            print(f"       isPK = obj['isPK'], isMultiple = {obj['isMultiple']}, isWeak = {obj['isWeak']}, composedOf = {obj['composedOf']}")
+
+    def _add_obj_is_a(self, obj):
+        self.add_is_a(
+            superClassLabel = obj['superClassLabel'], subclassParam = json.loads(obj['subClasses']),
+            superLabel = obj['superLabel'], subLabel = obj['subLabel'], isDisjunct = obj['isDisjunct']
+        )
+        if self.debug:
+            print(f"    ✓  added isA {json.loads(obj['subClasses'])} to super class {obj['superClassLabel']}:")
+            print(f"       superLabel = {obj['superLabel']}, subLabel = {obj['subLabel']}, isDisjunct = {obj['isDisjunct']}")
+
+    def _add_obj_rel(self, obj):
+        self.add_relation(
+            fromNodeLabel=obj['relationFrom'], 
+            relationLabel=obj['relationLabel'], 
+            toNodeLabel=obj['relationTo'], 
+            fromEdgeLabel=obj['fromEdgeLabel'], 
+            toEdgeLabel=obj['toEdgeLabel'], 
+            isWeak=obj['isWeak']
+        )
+        if self.debug:
+            print(f"    ✓  added relation {obj['relationLabel']} between {obj['relationFrom']} and {obj['relationTo']}:")
+            print(f"       fromEdgeLabel = {obj['fromEdgeLabel']}, toEdgeLabel = {obj['toEdgeLabel']}, isWeak = {obj['isWeak']}")
+
+    def __add_obj_copy(self, argument):
+        if self.has_obj(argument['label']):
+            if self.debug:
+                print(f" ! trying to add object {argument['label']} of type {argument['nodeType']}, but already exists")
+            return
+        switcher = {
+            str(NodeType.NODE): "_add_obj_node",
+            str(NodeType.ATTRIBUTE): "_add_obj_attr",
+            str(NodeType.IS_A): "_add_obj_is_a",
+            str(NodeType.RELATION): "_add_obj_rel"
+        }
+        function_name = switcher.get(argument['nodeType'], "not found")
+        if function_name == "not found":
+            raise Exception(f"  no adequate function found to add object copy for type {argument['nodeType']}")
+        #if self.debug:
+        #    print(f"    calling function {function_name}() to add object copy of type {argument['nodeType']}")
+        func = getattr(self, function_name, lambda: "Invalid node type")(argument)
+        return func
+
+    def mergeGraphsWith(self, otherGraph):
+        if self.debug:
+            print(" ")
+            print("|-> merging graphs")
+        for n1 in otherGraph.get_obj():
+            if self.has_obj(n1["label"]):
+                continue
+            else:
+                if self.debug:
+                    print(f" >  object '{n1['label']}' doesn't exist in this graph")
+            if (n1['nodeType'] == str(NodeType.COMPOSED_ATTRIBUTE)):
+                if self.debug:
+                    print(f" -> object copy of {NodeType.COMPOSED_ATTRIBUTE}s not supported directly, copy via parent object instead")
+                continue
+            
+            if self.debug:
+                print(f" » checking object '{n1['label']}' of type {n1['nodeType']}")
+            self.__add_obj_copy(n1)
+            if self.debug: print(" ")
+
     def compareNodes(self, otherGraph, label = "", node_type = NodeType.NOT_SPECIFIED, scores={}):
         dist = 0
         if self.debug:
             print(" ¶ calculating distances between graphs:    ")
+        if scores == {}:
+            if self.debug:
+                print("  no scores provided, fallback to default.")
+            scores = self.get_default_scores()
         for n1 in self.get_obj(label, node_type = node_type):
             if self.debug:
                 print(f" » testing {n1['label']}:")
@@ -485,11 +593,14 @@ class ER:
     def has_attr(self, label):
         return self.has_obj(label, NodeType.ATTRIBUTE)
 
+    def has_comp_attr(self, label):
+        return self.has_obj(label, NodeType.COMPOSED_ATTRIBUTE)
+
     def has_rel(self, label):
         return self.has_obj(label, NodeType.RELATION)
 
     def has_isA(self, label):
-        return self.has_obj(label, NodeType.ATTRIBUTE)
+        return self.has_obj(label, NodeType.IS_A)
 
     def has_obj(self, label, node_type=NodeType.NOT_SPECIFIED):
         """"
